/**
 * Glide 2.x FPS Overlay -- Version 2018-09-30
 * (c) 2018 Thomas Perl <m@thp.io>
 *
 * Tested with:
 * - Glide wrapper: dgVoodoo
 * - Games:
 *   - Wing Commander Prophecy 3Dfx Test (wcp3dfx.exe)
 *   - Need For Speed 2 SE Demo (nfs2sea.exe)
 **/

#include <stdio.h>
#include <Windows.h>

// Redirect the definitions for Glide APIs that we override (DLL import vs export)
#define grBufferSwap gxBufferSwap
#define grDepthBufferMode gxDepthBufferMode
#define grConstantColorValue gxConstantColorValue
#define grChromakeyMode gxChromakeyMode
#define grClipWindow gxClipWindow
#define grAlphaTestFunction gxAlphaTestFunction
#define grAlphaBlendFunction gxAlphaBlendFunction
#define grColorCombine gxColorCombine

#define __MSC__
#include "glide.h"

// Undo name redirections, s we can define the functions below
#undef grBufferSwap
#undef grDepthBufferMode
#undef grConstantColorValue
#undef grChromakeyMode
#undef grClipWindow
#undef grAlphaTestFunction
#undef grAlphaBlendFunction
#undef grColorCombine

// Forward all other calls (exported symbols) to the "glide2y.dll" library
#pragma comment(linker, "/def:glidefps.def")

typedef void (_stdcall * grBufferSwap_t)(int);
typedef void (_stdcall * grDepthBufferMode_t)(GrDepthBufferMode_t);
typedef void (_stdcall * grConstantColorValue_t)(GrColor_t);
typedef void (_stdcall * grChromakeyMode_t)(GrChromakeyMode_t);
typedef void (_stdcall * grClipWindow_t)(FxU32, FxU32, FxU32, FxU32);
typedef void (_stdcall * grAlphaTestFunction_t)(GrCmpFnc_t);
typedef void (_stdcall * grAlphaBlendFunction_t)(GrAlphaBlendFnc_t, GrAlphaBlendFnc_t, GrAlphaBlendFnc_t, GrAlphaBlendFnc_t);
typedef void (_stdcall * grColorCombine_t)(GrCombineFunction_t, GrCombineFactor_t, GrCombineLocal_t, GrCombineOther_t, FxBool);

// Dynamically loaded "glide2y.dll" and function pointers
static struct {
	HMODULE dll;

    grBufferSwap_t grBufferSwap;
    grDepthBufferMode_t grDepthBufferMode;
    grConstantColorValue_t grConstantColorValue;
    grChromakeyMode_t grChromakeyMode;
	grClipWindow_t grClipWindow;
	grAlphaTestFunction_t grAlphaTestFunction;
	grAlphaBlendFunction_t grAlphaBlendFunction;
	grColorCombine_t grColorCombine;
} glide2y = {0};

// Saved/tracked Glide state that we overwrite and need to restore
struct SavedGlideState {
	GrDepthBufferMode_t depthBufferMode;
	GrColor_t constantColorValue;
	GrChromakeyMode_t chromakeyMode;
	struct {
		FxU32 minx;
		FxU32 miny;
		FxU32 maxx;
		FxU32 maxy;
	} clipWindow;
	GrCmpFnc_t alphaTestFunction;
	struct {
		GrAlphaBlendFnc_t rgb_sf;
		GrAlphaBlendFnc_t rgb_df;
        GrAlphaBlendFnc_t alpha_sf;
		GrAlphaBlendFnc_t alpha_df;
	} alpheBlendFunction;
	struct {
		GrCombineFunction_t function;
		GrCombineFactor_t factor;
        GrCombineLocal_t local;
		GrCombineOther_t other;
        FxBool invert;
	} colorCombine;
};

static SavedGlideState
g_savedGlideState = {
	GR_DEPTHBUFFER_DISABLE,
	0xFFFFFFFF,
	GR_CHROMAKEY_DISABLE,
	{ 0, 0, 1024, 768, },
    GR_CMP_ALWAYS,
	{ GR_BLEND_ONE, GR_BLEND_ZERO, GR_BLEND_ONE, GR_BLEND_ZERO },
	{ GR_COMBINE_FUNCTION_SCALE_OTHER, GR_COMBINE_FACTOR_ONE, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_ITERATED, FXFALSE, },
};

static const char *
info = "@(#)Glide2x FPS Overlay -- Version 2018-09-30 Thomas Perl <m@thp.io>";

/**
 * Digit Definition
 * For the digits, imagine a 3x3 grid, with 0 = left, 1 = middle, 2 = right (and 0 = top,
 * 1 = middle, 2 = bottom). Each digit then consists of lines in this grid, the grid is
 * later scaled to the right size, this way each line can fit in a 16-bit value (4x3 bits)
 *  _      _
 * | |  |  _|
 * |_|  | |_  ...
 **/

struct DigitLine {
	unsigned short x0:3, y0:3, x1:3, y1:3;
};

struct Digit {
	unsigned char nlines;
	DigitLine lines[5];
};

static const struct Digit
digits[10] = {
	{ 4, { {0, 0, 2, 0}, {0, 2, 2, 2}, {0, 0, 0, 2}, {2, 0, 2, 2}, }, },
	{ 1, { {1, 0, 1, 2}, }, },
	{ 5, { {0, 0, 2, 0}, {0, 1, 2, 1}, {0, 2, 2, 2}, {2, 0, 2, 1}, {0, 1, 0, 2}, }, },
	{ 4, { {0, 0, 2, 0}, {0, 1, 2, 1}, {0, 2, 2, 2}, {2, 0, 2, 2}, }, },
	{ 3, { {0, 0, 0, 1}, {0, 1, 2, 1}, {2, 0, 2, 2}, }, },
	{ 5, { {0, 0, 2, 0}, {0, 1, 2, 1}, {0, 2, 2, 2}, {0, 0, 0, 1}, {2, 1, 2, 2}, }, },
	{ 4, { {0, 1, 2, 1}, {0, 2, 2, 2}, {0, 0, 0, 2}, {2, 1, 2, 2}, }, },
	{ 2, { {0, 0, 2, 0}, {2, 0, 2, 2}, }, },
	{ 5, { {0, 0, 2, 0}, {0, 1, 2, 1}, {0, 2, 2, 2}, {0, 0, 0, 2}, {2, 0, 2, 2}, }, },
	{ 5, { {0, 0, 2, 0}, {0, 1, 2, 1}, {0, 2, 2, 2}, {0, 0, 0, 1}, {2, 0, 2, 2}, }, },
};

static void
render_number(float x, float y, unsigned int value)
{
	GrVertex a, b;

	a.ooz = b.ooz = 65535.f;
	a.oow = b.oow = 1.f;

	a.a = b.a = 255.0f;
	a.r = b.r = 255.0f;
	a.g = b.g = 255.0f;
	a.b = b.b = 255.0f;

#define FONT_WIDTH 8
#define FONT_HEIGHT 16
#define FONT_STEP 10

	// Allow for 2 characters
	x += FONT_STEP;

	do {
		int cval = value % 10;

		const Digit &digit = digits[cval];
		for (int i=0; i<digit.nlines; ++i) {
			const DigitLine &line = digit.lines[i];
			a.x = x + FONT_WIDTH * (float)line.x0/2.f;
			b.x = x + FONT_WIDTH * (float)line.x1/2.f;
			a.y = y + FONT_HEIGHT * (float)line.y0/2.f;
			b.y = y + FONT_HEIGHT * (float)line.y1/2.f;
			grDrawLine(&a, &b);
		}

		x -= FONT_STEP;

		value /= 10;
	} while (value != 0);
}

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
			glide2y.dll = LoadLibrary("glide2y.dll");
			glide2y.grBufferSwap = (grBufferSwap_t)GetProcAddress(glide2y.dll, "_grBufferSwap@4");
			glide2y.grDepthBufferMode = (grDepthBufferMode_t)GetProcAddress(glide2y.dll, "_grDepthBufferMode@4");
			glide2y.grConstantColorValue = (grConstantColorValue_t)GetProcAddress(glide2y.dll, "_grConstantColorValue@4");
			glide2y.grChromakeyMode = (grChromakeyMode_t)GetProcAddress(glide2y.dll, "_grChromakeyMode@4");
			glide2y.grClipWindow = (grClipWindow_t)GetProcAddress(glide2y.dll, "_grClipWindow@16");
			glide2y.grAlphaTestFunction = (grAlphaTestFunction_t)GetProcAddress(glide2y.dll, "_grAlphaTestFunction@4");
			glide2y.grAlphaBlendFunction = (grAlphaBlendFunction_t)GetProcAddress(glide2y.dll, "_grAlphaBlendFunction@16");
			glide2y.grColorCombine = (grColorCombine_t)GetProcAddress(glide2y.dll, "_grColorCombine@20");
			break;
		case DLL_THREAD_ATTACH:
			break;
		case DLL_THREAD_DETACH:
			break;
		case DLL_PROCESS_DETACH:
			FreeLibrary(glide2y.dll), glide2y.dll = 0;
			break;
    }
    return TRUE;
}

extern "C" __declspec(dllexport) void _stdcall
grDepthBufferMode(GrDepthBufferMode_t mode)
{
	g_savedGlideState.depthBufferMode = mode;
	glide2y.grDepthBufferMode(mode);
}

extern "C" __declspec(dllexport) void _stdcall
grConstantColorValue(GrColor_t value)
{
	g_savedGlideState.constantColorValue = value;
	glide2y.grConstantColorValue(value);
}

extern "C" __declspec(dllexport) void _stdcall
grChromakeyMode(GrChromakeyMode_t mode)
{
	g_savedGlideState.chromakeyMode = mode;
	glide2y.grChromakeyMode(mode);
}

extern "C" __declspec(dllexport) void _stdcall
grClipWindow( FxU32 minx, FxU32 miny, FxU32 maxx, FxU32 maxy )
{
	g_savedGlideState.clipWindow.minx = minx;
	g_savedGlideState.clipWindow.miny = miny;
	g_savedGlideState.clipWindow.maxx = maxx;
	g_savedGlideState.clipWindow.maxy = maxy;
	glide2y.grClipWindow(minx, miny, maxx, maxy);
}

extern "C" __declspec(dllexport) void _stdcall
grAlphaTestFunction( GrCmpFnc_t function )
{
	g_savedGlideState.alphaTestFunction = function;
	glide2y.grAlphaTestFunction(function);
}

extern "C" __declspec(dllexport) void _stdcall
grAlphaBlendFunction(GrAlphaBlendFnc_t rgb_sf,   GrAlphaBlendFnc_t rgb_df,
                     GrAlphaBlendFnc_t alpha_sf, GrAlphaBlendFnc_t alpha_df)
{
	g_savedGlideState.alpheBlendFunction.rgb_sf = rgb_sf;
	g_savedGlideState.alpheBlendFunction.rgb_df = rgb_df;
	g_savedGlideState.alpheBlendFunction.alpha_sf = alpha_sf;
	g_savedGlideState.alpheBlendFunction.alpha_df = alpha_df;
	glide2y.grAlphaBlendFunction(rgb_sf, rgb_df, alpha_sf, alpha_df);
}

extern "C" __declspec(dllexport) void _stdcall
grColorCombine(GrCombineFunction_t function, GrCombineFactor_t factor,
               GrCombineLocal_t local, GrCombineOther_t other,
               FxBool invert)
{
	g_savedGlideState.colorCombine.function = function;
	g_savedGlideState.colorCombine.factor = factor;
	g_savedGlideState.colorCombine.local = local;
	g_savedGlideState.colorCombine.other = other;
	g_savedGlideState.colorCombine.invert = invert;
	glide2y.grColorCombine(function, factor, local, other, invert);
}

extern "C" __declspec(dllexport) void _stdcall
grBufferSwap(int swap_interval)
{
	static void *ptr = (void *)&grGlideInit;

	static DWORD started = 0,
				 frames = 0,
				 fps = 0;

	if (started == 0) {
		started = GetTickCount();
	}

	++frames;

	DWORD now = GetTickCount();
	if ((now - started) > 1000) {
		fps = frames;
		frames = 0;
		started = now;
	}

	// Set known-good render state for rendering the overlay
	glide2y.grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER, GR_COMBINE_FACTOR_ONE,
		GR_COMBINE_LOCAL_NONE, GR_COMBINE_OTHER_CONSTANT, FXFALSE);
	glide2y.grDepthBufferMode(GR_DEPTHBUFFER_DISABLE);
	glide2y.grChromakeyMode(GR_CHROMAKEY_DISABLE);
	glide2y.grClipWindow(0, 0, 640, 480);
	glide2y.grAlphaTestFunction(GR_CMP_ALWAYS);
	glide2y.grAlphaBlendFunction(GR_BLEND_ONE, GR_BLEND_ZERO, GR_BLEND_ONE, GR_BLEND_ZERO);
	
	//grFogMode(GR_FOG_DISABLE);
	//grColorMask(FXTRUE, FXTRUE);

	// Render the overlay
	glide2y.grConstantColorValue(0x00000000);
    render_number(15.f, 15.f, fps);
	render_number(13.f, 13.f, fps);
	glide2y.grConstantColorValue(0xFFFFFFFF);
    render_number(14.f, 14.f, fps);

	// Do the real swap
	glide2y.grBufferSwap(swap_interval);

	// Restore state
	glide2y.grDepthBufferMode(g_savedGlideState.depthBufferMode);
	glide2y.grConstantColorValue(g_savedGlideState.constantColorValue);
	glide2y.grChromakeyMode(g_savedGlideState.chromakeyMode);
	glide2y.grClipWindow(g_savedGlideState.clipWindow.minx,
		                 g_savedGlideState.clipWindow.miny,
						 g_savedGlideState.clipWindow.maxx,
						 g_savedGlideState.clipWindow.maxy);
	glide2y.grAlphaTestFunction(g_savedGlideState.alphaTestFunction);
	glide2y.grAlphaBlendFunction(g_savedGlideState.alpheBlendFunction.rgb_sf,
								 g_savedGlideState.alpheBlendFunction.rgb_df, 
							     g_savedGlideState.alpheBlendFunction.alpha_sf,
								 g_savedGlideState.alpheBlendFunction.alpha_df);
	glide2y.grColorCombine(g_savedGlideState.colorCombine.function,
						   g_savedGlideState.colorCombine.factor,
						   g_savedGlideState.colorCombine.local,
						   g_savedGlideState.colorCombine.other,
						   g_savedGlideState.colorCombine.invert);
}
